﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using symdump.symfile.util;
using symdump.util;

namespace symdump.symfile
{
    public class SymFile
    {
        private readonly Dictionary<string, EnumDef> _enums = new Dictionary<string, EnumDef>();
        //private readonly SortedSet<string> _externs = new SortedSet<string>();
        private List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration> _externs = new List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration>();
        public readonly List<Function> Functions = new List<Function>();
        private readonly Dictionary<string, string> _funcTypes = new Dictionary<string, string>();
        internal readonly Dictionary<uint, List<Label>> Labels = new Dictionary<uint, List<Label>>();
        private readonly Dictionary<string, StructDef> _structs = new Dictionary<string, StructDef>();
        private readonly byte _targetUnit;
        private readonly Dictionary<string, TypeInfo> _typedefs = new Dictionary<string, TypeInfo>();
        private readonly Dictionary<string, UnionDef> _unions = new Dictionary<string, UnionDef>();
        private readonly byte _version;
        private string _mxInfo;
        List<BenLincoln.ExecutableFiles.MemorySegment> MemorySegments { get; set; }
        public uint ProgramCounter { get; set; }
        public uint GlobalPointer { get; set; }
        public uint StackPointerBase { get; set; }
        public uint StackPointerSizeOrOffset { get; set; }

        public SymFile(BinaryReader stream)
        {
            MemorySegments = new List<BenLincoln.ExecutableFiles.MemorySegment> { };
            stream.BaseStream.Seek(0, SeekOrigin.Begin);

            stream.Skip(3);
            _version = stream.ReadByte();
            _targetUnit = stream.ReadByte();

            stream.Skip(3);

            BenLincoln.DebugSymbols.SequenceIssuer issuer = new BenLincoln.DebugSymbols.SequenceIssuer();

            while (stream.BaseStream.Position < stream.BaseStream.Length)
            {
                DumpEntry(stream, issuer);
            }
        }

        public void GetDataFromPSXEXE(string psxEXEPath)
        {
            BenLincoln.ExecutableFiles.BinaryFile inputFile = new BenLincoln.ExecutableFiles.BinaryFile();
            BenLincoln.ExecutableFiles.PSXEXE inputFilePSXEXE = new BenLincoln.ExecutableFiles.PSXEXE();
            string exceptionMessagePSXEXE = "";

            try
            {
                FileStream inputStream = new FileStream(psxEXEPath, FileMode.Open, FileAccess.Read);
                BinaryReader inputReader = new BinaryReader(inputStream);

                try
                {
                    inputFilePSXEXE = new BenLincoln.ExecutableFiles.PSXEXE();
                    inputFilePSXEXE.ReadFile(inputReader);
                    
                    int segNum = 1;

                    // text segment
                    BenLincoln.ExecutableFiles.MemorySegment msText = new BenLincoln.ExecutableFiles.MemorySegment();
                    msText.SegmentNumber = segNum;
                    msText.Name = "text";
                    msText.Offset = inputFilePSXEXE.PSXExec.TextSegmentAddress;
                    msText.Length = inputFilePSXEXE.PSXExec.TextSegmentSize;
                    MemorySegments.Add(msText);
                    segNum++;

                    ProgramCounter = inputFilePSXEXE.PSXExec.ProgramCounter;
                    GlobalPointer = inputFilePSXEXE.PSXExec.GlobalPointer;
                    StackPointerBase = inputFilePSXEXE.PSXExec.StackSegmentAddress;
                    StackPointerSizeOrOffset = inputFilePSXEXE.PSXExec.StackSegmentSize;

                    // segments that usually don't exist
                    if (inputFilePSXEXE.PSXExec.DataSegmentAddress != 0)
                    {
                        BenLincoln.ExecutableFiles.MemorySegment msData = new BenLincoln.ExecutableFiles.MemorySegment();
                        msData.SegmentNumber = segNum;
                        msData.Name = "data";
                        msData.Offset = inputFilePSXEXE.PSXExec.DataSegmentAddress;
                        msData.Length = inputFilePSXEXE.PSXExec.DataSegmentSize;
                        MemorySegments.Add(msData);
                        segNum++;
                    }

                    if (inputFilePSXEXE.PSXExec.BSSSegmentAddress != 0)
                    {
                        BenLincoln.ExecutableFiles.MemorySegment msBSS = new BenLincoln.ExecutableFiles.MemorySegment();
                        msBSS.SegmentNumber = segNum;
                        msBSS.Name = "bss";
                        msBSS.Offset = inputFilePSXEXE.PSXExec.BSSSegmentAddress;
                        msBSS.Length = inputFilePSXEXE.PSXExec.BSSSegmentSize;
                        MemorySegments.Add(msBSS);
                        segNum++;
                    }

                    // RAM
                    BenLincoln.ExecutableFiles.MemorySegment msStack = new BenLincoln.ExecutableFiles.MemorySegment();
                    msStack.SegmentNumber = segNum;
                    msStack.Name = "ram";
                    //msStack.Offset = 0x801F8000;
                    msStack.Offset = inputFilePSXEXE.PSXExec.TextSegmentAddress + inputFilePSXEXE.PSXExec.TextSegmentSize;
                    if (msStack.Offset > 0x80200000)
                    {
                        throw new Exception(String.Format("Unsupported configuration: text segment address + length (0x{0:X})is greater than 0x80200000", msStack.Offset));
                    }
                    uint ramEndAddress = 0x80200000;
                    // this is a hack, sorry
                    if (msStack.Offset < 0x00200000)
                    {
                        ramEndAddress = 0x00200000;
                    }
                    //else
                    //{
                    //    throw new Exception(String.Format("Unsupported configuration: couldn't calculate valid RAM end address for stack offset 0x{0:X8}", msStack.Offset));
                    //}
                    msStack.Length = ramEndAddress - msStack.Offset;
                    MemorySegments.Add(msStack);
                    segNum++;

                    // RAM

                }
                catch (Exception fe)
                {
                    exceptionMessagePSXEXE = fe.Message;
                }

            }
            catch (Exception ex)
            {
                throw new Exception(String.Format("Could not read the input file '{0}': {1}", psxEXEPath, ex.Message));
            }
        }

        public void Dump(TextWriter output)
        {
            var writer = new IndentedTextWriter(output);
            writer.WriteLine($"Version = {_version}, targetUnit = {_targetUnit}");

            writer.WriteLine();
            writer.WriteLine($"// {_enums.Count} enums");
            foreach (var e in _enums.Values)
                e.Dump(writer);

            writer.WriteLine();
            writer.WriteLine($"// {_unions.Count} unions");
            foreach (var e in _unions.Values)
                e.Dump(writer);

            writer.WriteLine();
            writer.WriteLine($"// {_structs.Count} structs");
            foreach (var e in _structs.Values)
                e.Dump(writer);

            writer.WriteLine();
            writer.WriteLine($"// {_typedefs.Count} typedefs");
            foreach (var t in _typedefs)
                writer.WriteLine($"typedef {t.Value.AsCode(t.Key)};");

            writer.WriteLine();
            writer.WriteLine($"// {Labels.Count} labels");
            foreach (var l in Labels)
            foreach (var l2 in l.Value)
                writer.WriteLine(l2);

            writer.WriteLine();
            writer.WriteLine($"// {_externs.Count} external declarations");
            foreach (var e in _externs)
                writer.WriteLine(e);

            writer.WriteLine();
            writer.WriteLine($"// {Functions.Count} functions");
            foreach (var f in Functions)
                f.Dump(writer);
        }

        public BenLincoln.DebugSymbols.SymDump.SymDumpSymbolCollection ToBLDSSymDumpSymbolCollection(string defaultSourceFileNameBase, bool autoRenameFakes)
        {
            BenLincoln.DebugSymbols.SymDump.SymDumpSymbolCollection result = new BenLincoln.DebugSymbols.SymDump.SymDumpSymbolCollection();

            string defaultSourceFileHeader = String.Format("{0}.H", defaultSourceFileNameBase);
            string defaultSourceFileCode = String.Format("{0}.C", defaultSourceFileNameBase);

            result.version = _version;
            result.target_unit = _targetUnit;
            result.program_counter = ProgramCounter;
            result.global_pointer = GlobalPointer;
            result.stack_pointer_base = StackPointerBase;
            result.stack_pointer_size_or_offset = StackPointerSizeOrOffset;

            Dictionary<int, StructDef> structCollection = new Dictionary<int, StructDef>();
            Dictionary<int, UnionDef> unionCollection = new Dictionary<int, UnionDef>();
            List<int> structSequenceNumbers = new List<int>();
            List<int> unionSequenceNumbers = new List<int>();
            foreach (string k in _structs.Keys)
            {
                if (structSequenceNumbers.Contains(_structs[k].SYMFileSequenceNumber))
                {
                    throw new Exception(string.Format("List of struct SYM file sequence numbers already contains {0}", (_structs[k].SYMFileSequenceNumber)));
                }
                structSequenceNumbers.Add(_structs[k].SYMFileSequenceNumber);
                structCollection.Add(_structs[k].SYMFileSequenceNumber, _structs[k]);
                //result.structs.Add(s);
            }
            structSequenceNumbers.Sort();
            foreach (string k in _unions.Keys)
            {
                //BenLincoln.DebugSymbols.SymDump.CUnion u = _unions[k].ToBLDSCUnion();
                //result.unions.Add(u);
                if (unionSequenceNumbers.Contains(_unions[k].SYMFileSequenceNumber))
                {
                    throw new Exception(string.Format("List of union SYM file sequence numbers already contains {0}", (_unions[k].SYMFileSequenceNumber)));
                }
                unionSequenceNumbers.Add(_unions[k].SYMFileSequenceNumber);
                unionCollection.Add(_unions[k].SYMFileSequenceNumber, _unions[k]);
            }
            unionSequenceNumbers.Sort();
            if (autoRenameFakes)
            {
                Dictionary<string, string> structNameMap = new Dictionary<string, string>();
                Dictionary<string, string> unionNameMap = new Dictionary<string, string>();

                foreach (string k in _structs.Keys)
                {
                    //Console.WriteLine(k);
                    //_structs[k].Members[0];
                }
                //for (int i = 0; i < result.structs.Count; i++)
                //{

                //}
            }

            result.structs = new List<BenLincoln.DebugSymbols.SymDump.CStruct>();
            for (int i = 0; i < structSequenceNumbers.Count; i++)
            {
                result.structs.Add(structCollection[structSequenceNumbers[i]].ToBLDSCStruct());
            }

            result.unions = new List<BenLincoln.DebugSymbols.SymDump.CUnion>();
            for (int i = 0; i < unionSequenceNumbers.Count; i++)
            {
                result.unions.Add(unionCollection[unionSequenceNumbers[i]].ToBLDSCUnion());
            }

            // directly from EXE
            result.memory_segments = new List<BenLincoln.DebugSymbols.SymDump.MemorySegment> { };
            if (MemorySegments.Count > 0)
            {
                for (int i = 0; i < MemorySegments.Count; i++)
                {
                    BenLincoln.DebugSymbols.SymDump.MemorySegment ms = new BenLincoln.DebugSymbols.SymDump.MemorySegment();
                    ms.name = MemorySegments[i].Name;
                    ms.offset = MemorySegments[i].Offset;
                    ms.length = MemorySegments[i].Length;
                    ms.segment_number = MemorySegments[i].SegmentNumber;
                    result.memory_segments.Add(ms);
                }
            }

            result.external_declarations = new List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration>();
            foreach (BenLincoln.DebugSymbols.SymDump.CExternalDeclaration x in _externs)
            {
                result.external_declarations.Add(x);
            }

            result.enums = new List<BenLincoln.DebugSymbols.SymDump.CEnum>();
            foreach (string k in _enums.Keys)
            {
                EnumDef oe = _enums[k];
                BenLincoln.DebugSymbols.SymDump.CEnum e = oe.ToBLDSEnumDef(defaultSourceFileHeader);
                result.enums.Add(e);
            }

            result.typedefs = new List<BenLincoln.DebugSymbols.SymDump.CTypeDef>();
            foreach (string k in _typedefs.Keys)
            {
                BenLincoln.DebugSymbols.SymDump.CTypeDef t = _typedefs[k].ToBLDSCTypeDef(k, defaultSourceFileHeader);
                // redefine pointer typedefs correctly
                if (t.name.StartsWith("*"))
                {
                    t.name = t.name.Substring(1);
                    t.c_type = String.Format("{0} *", t.c_type);
                }
                // redefine function pointer typedefs correctly (I think?)
                if (t.name.StartsWith("(*"))
                {
                    t.name = t.name.Substring(2);
                    int firstClosingParenPos = t.name.IndexOf(")");
                    t.name = t.name.Substring(0, firstClosingParenPos);
                    t.c_type = "void *";
                }

                result.typedefs.Add(t);
            }

            result.labels = new List<BenLincoln.DebugSymbols.SymDump.CLabel>();
            foreach (uint k in Labels.Keys)
            {
                List<Label> labelList = Labels[k];
                foreach (Label l in labelList)
                {
                    result.labels.Add(l.ToBLDSCLabel());
                }
            }

            result.functions = new List<BenLincoln.DebugSymbols.SymDump.CFunction>();
            foreach (Function f in Functions)
            {
                result.functions.Add(f.ToBLDSCFunction(defaultSourceFileCode));
            }

            return result;
        }

        public string DumpJSON(string defaultSourceFileNameBase, bool autoRenameFakes)
        {
            BenLincoln.DebugSymbols.SymDump.SymDumpSymbolCollection resultDef = ToBLDSSymDumpSymbolCollection(defaultSourceFileNameBase, autoRenameFakes);

            return resultDef.ToJSON();
        }

        private void DumpEntry(BinaryReader stream, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            var typedValue = new TypedValue(stream, issuer);
            if (typedValue.Type == 8)
            {
                _mxInfo = $"${typedValue.Value:X} MX-info {stream.ReadByte():X}";
                return;
            }

            if (typedValue.IsLabel)
            {
                try
                {
                    var lbl = new Label(typedValue, stream);

                    if (!Labels.ContainsKey(lbl.Offset))
                    {
                        Labels.Add(lbl.Offset, new List<Label>());
                    }

                    Labels[lbl.Offset].Add(lbl);
                }
                catch (Exception e)
                {
                    Console.WriteLine(string.Format("Exception thrown while reading label data: {0}", e.Message));
                }
                return;
            }

            int tvt = typedValue.Type & 0x7f;
            switch (tvt)
            {
                case 0:
#if WITH_SLD
                writer.WriteLine($"${typedValue.value:X} Inc SLD linenum");
                #endif
                    break;
                case 2:
#if WITH_SLD
                writer.WriteLine($"${typedValue.value:X} Inc SLD linenum by byte {stream.ReadU1()}");
                #else
                    stream.Skip(1);
#endif
                    break;
                case 4:
#if WITH_SLD
                writer.WriteLine($"${typedValue.value:X} Inc SLD linenum by word {stream.ReadUInt16()}");
#else
                    stream.Skip(2);
#endif
                    break;
                case 6:
#if WITH_SLD
                writer.WriteLine($"${typedValue.value:X} Set SLD linenum to {stream.ReadUInt32()}");
#else
                    stream.Skip(4);
#endif
                    break;
                case 8:
#if WITH_SLD
                writer.WriteLine($"${typedValue.value:X} Set SLD to line {stream.ReadUInt32()} of file " +
                    stream.readPascalString());
#else
                    stream.Skip(4);
                    stream.Skip(stream.ReadByte());
#endif
                    break;
                case 10:
#if WITH_SLD
                writer.WriteLine($"${typedValue.value:X} End SLD info");
#endif
                    break;
                case 12:
                    DumpType12(stream, typedValue.Value, issuer);
                    break;
                case 20:
                    DumpType20(stream, typedValue.Value, issuer);
                    break;
                case 22:
                    DumpType22(stream, typedValue.Value, issuer);
                    break;
                default:
                    //throw new Exception("Sodom");
                    Console.WriteLine(string.Format("Unhandled type number: {0} - data will be ignored", tvt));
                    break;
            }
        }

        private void DumpType12(BinaryReader stream, int offset, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            Functions.Add(new Function(stream, (uint) offset, _funcTypes, issuer));
            //writer.WriteLine("{");
            //++writer.Indent;
        }

        private void ReadEnum(BinaryReader reader, string name, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            var e = new EnumDef(reader, name, issuer);

            EnumDef already;
            if (_enums.TryGetValue(name, out already))
            {
                //                if (!e.Equals(already))
                //                throw new Exception($"Non-uniform definitions of enum {name}");
                //                return;
                if (e.Equals(already))
                {
                    return;
                }
                else
                {
                    string originalName = name;
                    int iterationNum = 1;
                    while (_enums.TryGetValue(name, out already))
                    {
                        if (e.Equals(already))
                        {
                            return;
                        }
                        name = String.Format("{0}_alt_definition_{1:D4}", originalName, iterationNum);
                        iterationNum++;
                    }
                    Console.WriteLine($"WARNING: Non-uniform definitions of enum {originalName}. The redefined version has been renamed to {name}.");
                    e.Name = name;
                }
            }

            _enums.Add(name, e);
        }

        private void ReadUnion(BinaryReader reader, string name, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            string unionName = name;
            //if (name.StartsWith("."))
            //{
            //    unionName = String.Format("_{0}", unionName.Substring(1));
            //}
            var e = new UnionDef(reader, unionName, issuer);

            UnionDef already;
            if (_unions.TryGetValue(unionName, out already))
            {
                if (e.Equals(already))
                {
                    Console.WriteLine(string.Format("Info: duplicate definition of union '{0}'. Existing definition has hashcode 0x{1:X8} (dec: {1}), and duplicate has hashcode 0x{2:X8} (dec: {2}).", name, e.GetHashCode(), already.GetHashCode()));
                    return;
                }

                Console.WriteLine(string.Format("Warning: redefinition of union '{0}' with identical name. Existing definition has hashcode 0x{1:X8} (dec: {1}), and duplicate has hashcode 0x{2:X8} (dec: {2}).", name, e.GetHashCode(), already.GetHashCode()));
        
                if (e.IsFake)
                {
                    // generate new "fake fake" name
                    var n = 0;
                    while (_unions.ContainsKey($"{unionName}_{n}"))
                        ++n;

                    _unions.Add($"{unionName}_{n}", e);
                    Console.WriteLine(string.Format("\tThis is a 'fake' union. Using '{0}' as its key", $"{name}_{n}"));

                    return;
                }
                else
                {
                    //throw new Exception($"Non-uniform definitions of union {name}");
                    string originalName = unionName;
                    int iterationNum = 1;
                    while (_unions.TryGetValue(unionName, out already))
                    {
                        if (e.Equals(already))
                        {
                            Console.WriteLine(string.Format("\tInfo: duplicate definition of union '{0}'. Existing definition has hashcode 0x{1:X8} (dec: {1}), and duplicate has hashcode 0x{2:X8} (dec: {2}).", name, e.GetHashCode(), already.GetHashCode()));
                            return;
                        }
                        unionName = String.Format("{0}_alt_definition_{1:D4}", originalName, iterationNum);
                        iterationNum++;
                    }
                    Console.WriteLine($"\tWARNING: Non-uniform definitions of union {originalName}. The redefined version has been given the key {unionName}.");
                    e.Name = unionName;
                }
            }

            _unions.Add(unionName, e);
        }

        private void ReadStruct(BinaryReader reader, string name, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            var e = new StructDef(reader, name, issuer);

            StructDef already;
            if (_structs.TryGetValue(name, out already))
            {
                if (e.Equals(already))
                {
                    Console.WriteLine(string.Format("Info: duplicate definition of struct '{0}'. Existing definition has hashcode 0x{1:X8} (dec: {1}), and duplicate has hashcode 0x{2:X8} (dec: {2}).", name, e.GetHashCode(), already.GetHashCode()));
                    return;
                }
                Console.WriteLine(string.Format("Warning: redefinition of struct '{0}' with identical name. Existing definition has hashcode 0x{1:X8} (dec: {1}), and duplicate has hashcode 0x{2:X8} (dec: {2}).", name, e.GetHashCode(), already.GetHashCode()));

                if (e.IsFake)
                {
                    // generate new "fake fake" name
                    var n = 0;
                    while (_structs.ContainsKey($"{name}_{n}"))
                        ++n;

                    _structs.Add($"{name}_{n}", e);
                    Console.WriteLine(string.Format("\tThis is a 'fake' struct. Using '{0}' as its key", $"{name}_{n}"));

                    return;
                }
                else
                {
                    // Console.WriteLine($"WARNING: Non-uniform definitions of struct {name}");
                    string originalName = name;
                    int iterationNum = 1;
                    while (_structs.TryGetValue(name, out already))
                    {
                        if (e.Equals(already))
                        {
                            Console.WriteLine(string.Format("\tInfo: duplicate definition of struct '{0}'. Existing definition has hashcode 0x{1:X8} (dec: {1}), and duplicate has hashcode 0x{2:X8} (dec: {2}).", name, e.GetHashCode(), already.GetHashCode()));
                            return;
                        }
                        name = String.Format("{0}_alt_definition_{1:D4}", originalName, iterationNum);
                        iterationNum++;
                    }
                    Console.WriteLine($"\tNon-uniform definitions of struct '{originalName}'. The redefined version has been given the key '{name}'.");
                    e.Name = name;
                }
            }

            _structs.Add(name, e);
        }

        private void AddTypedef(string name, TypeInfo typeInfo, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            TypeInfo already;
            if (_typedefs.TryGetValue(name, out already))
            {
                if (typeInfo.Equals(already))
                {
                    return;
                }
                else
                {
                    //throw new Exception($"Non-uniform definitions of typedef for {name}");
                    string originalName = name;
                    int iterationNum = 1;
                    while (_typedefs.TryGetValue(name, out already))
                    {
                        if (typeInfo.Equals(already))
                        {
                            return;
                        }
                        name = String.Format("{0}_alt_definition_{1:D4}", originalName, iterationNum);
                        iterationNum++;
                    }
                    Console.WriteLine($"WARNING: Non-uniform definitions of typedef for {originalName}. The redefined version has been renamed to {name}.");
                    // not sure if I need to rename the typedef somehow or not
                }
            }

            _typedefs.Add(name, typeInfo);
        }

        private void DumpType20(BinaryReader stream, int offset, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            var ti = stream.ReadTypeInfo(false, issuer);
            var name = stream.ReadPascalString();

            if (ti.ClassType == ClassType.Enum && ti.TypeDef.BaseType == BaseType.EnumDef)
            {
                ReadEnum(stream, name, issuer);
                return;
            }

            if (ti.ClassType == ClassType.FileName)
                return;
            if (ti.ClassType == ClassType.Struct && ti.TypeDef.BaseType == BaseType.StructDef)
                ReadStruct(stream, name, issuer);
            else if (ti.ClassType == ClassType.Union && ti.TypeDef.BaseType == BaseType.UnionDef)
                ReadUnion(stream, name, issuer);
            else if (ti.ClassType == ClassType.Typedef)
                AddTypedef(name, ti, issuer);
            else if (ti.ClassType == ClassType.External)
                if (ti.TypeDef.IsFunctionReturnType)
                {
                    _funcTypes[name] = ti.AsCode("").Trim();
                }
                else
                {
                    //_externs.Add($"extern {ti.AsCode(name)}; // offset 0x{offset:X}");
                    BenLincoln.DebugSymbols.SymDump.CExternalDeclaration x = new BenLincoln.DebugSymbols.SymDump.CExternalDeclaration();
                    x.declaration = $"extern {ti.AsCode(name)}; // offset 0x{offset:X}";
                    x.class_type = ti.ClassType.ToString();
                    x.offset = (uint)offset;
                    x.size = ti.Size;
                    x.extern_type = "extern";
                    _externs.Add(x);
                }
            else if (ti.ClassType == ClassType.Static)
                if (ti.TypeDef.IsFunctionReturnType)
                {
                    _funcTypes[name] = ti.AsCode("").Trim();
                }
                else
                {
                    //_externs.Add($"static {ti.AsCode(name)}; // offset 0x{offset:X}");
                    BenLincoln.DebugSymbols.SymDump.CExternalDeclaration x = new BenLincoln.DebugSymbols.SymDump.CExternalDeclaration();
                    x.declaration = $"static {ti.AsCode(name)}; // offset 0x{offset:X}";
                    x.class_type = ti.ClassType.ToString();
                    x.offset = (uint)offset;
                    x.size = ti.Size;
                    x.extern_type = "static";
                    _externs.Add(x);
                }
            else
                throw new Exception("Gomorrha");
        }

        private void DumpType22(BinaryReader stream, int offset, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            var ti = stream.ReadTypeInfo(true, issuer);
            var name = stream.ReadPascalString();

            if (ti.ClassType == ClassType.Enum && ti.TypeDef.BaseType == BaseType.EnumDef)
                ReadEnum(stream, name, issuer);
            else if (ti.ClassType == ClassType.Typedef)
                AddTypedef(name, ti, issuer);
            else if (ti.ClassType == ClassType.External)
                if (ti.TypeDef.IsFunctionReturnType)
                {
                    _funcTypes[name] = ti.AsCode("").Trim();
                }
                else
                {
                    //_externs.Add($"extern {ti.AsCode(name)}; // offset 0x{offset:X}");
                    BenLincoln.DebugSymbols.SymDump.CExternalDeclaration x = new BenLincoln.DebugSymbols.SymDump.CExternalDeclaration();
                    x.declaration = $"extern {ti.AsCode(name)}; // offset 0x{offset:X}";
                    x.class_type = ti.ClassType.ToString();
                    x.offset = (uint)offset;
                    x.size = ti.Size;
                    x.extern_type = "extern";
                    _externs.Add(x);
                }
            else if (ti.ClassType == ClassType.Static)
                if (ti.TypeDef.IsFunctionReturnType)
                {
                    _funcTypes[name] = ti.AsCode("").Trim();
                }
                else
                {
                    //_externs.Add($"static {ti.AsCode(name)}; // offset 0x{offset:X}");
                    BenLincoln.DebugSymbols.SymDump.CExternalDeclaration x = new BenLincoln.DebugSymbols.SymDump.CExternalDeclaration();
                    x.declaration = $"static {ti.AsCode(name)}; // offset 0x{offset:X}";
                    x.class_type = ti.ClassType.ToString();
                    x.offset = (uint)offset;
                    x.size = ti.Size;
                    x.extern_type = "static";
                    _externs.Add(x);
                }
            else
                throw new Exception("Gomorrha");
        }

        public Function FindFunction(uint addr)
        {
            return Functions.FirstOrDefault(f => f.Address == addr);
        }
    }
}
